package dk.dtu.caseid;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.StringTokenizer;

import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.GlobalContext;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMa;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMc;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMf;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMfa;
import org.processmining.plugins.InductiveMiner.plugins.IM;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.TreeFitnessAbstract;
import org.processmining.plugins.etm.fitness.metrics.FitnessReplay;
import org.processmining.plugins.etm.fitness.metrics.Generalization;
import org.processmining.plugins.etm.fitness.metrics.PrecisionEscEdgesImproved;
import org.processmining.plugins.etm.fitness.metrics.SimplicityMixed;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.processtree.ProcessTree;

import nl.tue.astar.AStarThread.Canceller;
import prompt.csv.data.CSVLog;
import prompt.csv.data.FIELD_ROLE;
import prompt.csv.utils.UnicodeBOMInputStream;
import prompt.csv.utils.context.DummyGlobalContext;
import prompt.csv.utils.context.DummyPluginContext;

public class CaseIdentifier {

	private GlobalContext context = new DummyGlobalContext();
	private PluginContext pluginContext = new DummyPluginContext(context, "");
	
	public void generateLogMetrics(String CSVFile, int timestampIndex, int activityIndex, int candidateCaseIdIndex, PrintStream out) throws Exception{
		
		CSVLog csv = read(CSVFile);
		ArrayList<String> colsname = getCandidateColumnsName(CSVFile);
		
		String[] algs = new String[] {"imfa"/*,"ima","imf","im","imc"*/}; // ,"imc" "imfa","ima","imf","im"
		
		for(String alg : algs){
			Map<TreeFitnessAbstract, Double> v = getValues(csv, timestampIndex, activityIndex, candidateCaseIdIndex, alg);
			//System.out.print("values obtained");
			//System.out.print(candiateCaseIds[i]);
			
			HashMap<String, Double> metricsTEMP = new HashMap<String, Double>();
			
//			System.out.println("");
			for(TreeFitnessAbstract tfa : v.keySet()) {
//				System.out.print(tfa.getInfo()+""+v.get(tfa)+" ,");
				metricsTEMP.put(tfa.getInfo().toString(), v.get(tfa));
			}
//			System.out.println("");
		
//			ArrayList<Double> metrics = new ArrayList<Double>();
//			metrics.add(metricsTEMP.get("Fr"));
//			metrics.add(metricsTEMP.get("Pi"));
//			metrics.add(metricsTEMP.get("Sm"));
//			metrics.add(metricsTEMP.get("Gv"));
			
			out.print("\"" + alg + "\"\t");
			out.print("\"" + CSVFile + "\"\t");
			out.print("\"" + colsname.get(candidateCaseIdIndex) + "\"\t");
			for (String m : new String[] {"Fr", "Pi", "Sm", "Gv"}) {
				out.print("\"" + m + "\"\t");
				out.print(metricsTEMP.get(m) + "\t");
			}
			out.println("");
			
			//System.out.print("in write to file");
//			writeTofile(outFile, CSVFile, colsname.get(candidateCaseIdIndex), metrics, alg);
			//System.out.print("out write to file");
			//System.out.println("");
		}
	}
	
	public ArrayList<String> getCandidateColumnsName(String filename) throws IOException{
		
		File file = new File(filename);
		FileReader fileReader = new FileReader(file);
		BufferedReader bufferedReader = new BufferedReader(fileReader);
		String line;
		line = bufferedReader.readLine().replace("\"", "") ;
		StringTokenizer st = new StringTokenizer(line,",");
		ArrayList<String> colsname = new ArrayList<String>();
		
		while(st.hasMoreElements()){
			colsname.add(st.nextToken());	
		}
		
		fileReader.close();
		
		return colsname ;
		
	}
	
	public static CSVLog read(String file) throws Exception {
		CSVLog log = new CSVLog(file);
		UnicodeBOMInputStream ubis = new UnicodeBOMInputStream(new FileInputStream(file));
		ubis.skipBOM();
		InputStreamReader isr = new InputStreamReader(ubis);
		BufferedReader reader = new BufferedReader(isr);
		String readLine;
		while ((readLine = reader.readLine()) != null) {
			log.addLine(readLine);
		}
		reader.close();
		isr.close();
		ubis.close();
		log.prepare();
		return log;
	}
	
	private Map<TreeFitnessAbstract, Double> getValues(CSVLog csv, int time, int activityName, int caseId,String alg) {
		//System.out.println("Xlog making starts");
		XLog log = getXLog(csv, time, activityName, caseId);
		// {"im","ima","imf","imfa","imc"}
		ProcessTree pt = null ;
		if(alg.equals("im"))  pt = this.doInductiveMinerIM(log);
		else if(alg.equals("ima"))  pt =this.doInductiveMinerIMa(log);
		else if(alg.equals("imf"))  pt =this.doInductiveMinerIMf(log);
		else if(alg.equals("imfa"))  pt =this.doInductiveMinerIMfa(log);
		else if(alg.equals("imc"))  pt =this.doInductiveMinerIMc(log);
		
	
		CentralRegistry registry = new CentralRegistry(log, new XEventNameClassifier(), new Random(1));
		ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree tree = converter.convert(pt);
		
		List<TreeFitnessAbstract> evals = new ArrayList<TreeFitnessAbstract>();
		evals.add(new FitnessReplay(registry,  new DummyCanceler())); // what is DummyCanceler ?
		//System.out.println("metric done1");
//		evals.add(new PrecisionEscEdges(registry));
		//System.out.println("metric done2");
		evals.add(new PrecisionEscEdgesImproved(registry));
		//System.out.println("metric done3");
//		evals.add(new PrecisionFlowerSize());
		//System.out.println("metric done4");
//		evals.add(new PrecisionOperatorCosts());
		//System.out.println("metric done5");
//		evals.add(new SimplicityDuplMissingAct(registry));
		//System.out.println("metric done6");
		evals.add(new SimplicityMixed());
		//System.out.println("metric done7");
//		evals.add(new SimplicityOperatorAlternation());
		//System.out.println("metric done8");
//		evals.add(new SimplicityTreeSize());
		//System.out.println("metric done9");
//		evals.add(new SimplicityUselessNodes());
		//System.out.println("metric done10");
		evals.add(new Generalization(registry));
		//System.out.println("metric done11");
//		evals.add(new GeneralizationByFitnessReplayDeviation(registry));
		//System.out.println("metric done12");
		
		// silent output
		//ByteArrayOutputStream pipeOut = new ByteArrayOutputStream();
		
		//PrintStream old_out = //System.out;
		
		//System.setOut(new PrintStream(pipeOut));
		
		
		Map<TreeFitnessAbstract, Double> values = new HashMap<TreeFitnessAbstract, Double>();
	
		for(TreeFitnessAbstract tfa : evals) {
			// bug here 
			//System.out.print("get fitness called");
			values.put(tfa, tfa.getFitness(tree, Arrays.asList(tree)));
			//System.out.print("get fitness over");
			System.gc();
		}
		
		// restore output
		//System.setOut(old_out);
		
		//System.out.println("return statement reached");
		
		return values;
	}
	
	private ProcessTree doInductiveMinerIM(XLog log) {
		//System.out.println("start inductive minner");
		IM im = new IM();
		MiningParametersIM para = new MiningParametersIM();
		return im.mineProcessTree(pluginContext, log, para); 
	}
	
	private ProcessTree doInductiveMinerIMf(XLog log) {
		//System.out.println("start inductive minner");
		IM im = new IM();
		MiningParametersIMf para = new MiningParametersIMf();
		
		return im.mineProcessTree(pluginContext, log, para); 
	
	
	}
	private ProcessTree doInductiveMinerIMfa(XLog log) {
		//System.out.println("start inductive minner");
		IM im = new IM();
		MiningParametersIMfa para = new MiningParametersIMfa();
		
		return im.mineProcessTree(pluginContext, log, para); 
	
	
	}
	private ProcessTree doInductiveMinerIMa(XLog log) {
		//System.out.println("start inductive minner");
		IM im = new IM();
		MiningParametersIMa para = new MiningParametersIMa();
		
		return im.mineProcessTree(pluginContext, log, para); 
	
	
	}
	private ProcessTree doInductiveMinerIMc(XLog log) {
		//System.out.println("start inductive minner");
		IM im = new IM();
		MiningParametersIMc para = new MiningParametersIMc();
		
		return im.mineProcessTree(pluginContext, log, para); 
	
	}
	private XLog getXLog(CSVLog csv, int time, int activityName, int caseId) {
		csv.setRole(time, FIELD_ROLE.DATE);
		csv.setRole(activityName, FIELD_ROLE.ACTIVITY_NAME);
		csv.setRole(caseId, FIELD_ROLE.CASE_ID);
		//System.out.println("to xlog starts");
		return csv.toXLog(null);
	}
}

class DummyCanceler implements Canceller {
	public boolean isCancelled() {
		return false;
	}
}
