package iciPlugin;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.StringTokenizer;

import nl.tue.astar.AStarThread.Canceller;

import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.plugin.GlobalContext;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIM;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMa;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMc;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMf;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMfa;
import org.processmining.plugins.InductiveMiner.plugins.IM;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.TreeFitnessAbstract;
import org.processmining.plugins.etm.fitness.metrics.FitnessReplay;
import org.processmining.plugins.etm.fitness.metrics.Generalization;
import org.processmining.plugins.etm.fitness.metrics.PrecisionEscEdgesImproved;
import org.processmining.plugins.etm.fitness.metrics.SimplicityMixed;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.processtree.ProcessTree;

import prompt.csv.data.CSVLog;
import prompt.csv.data.FIELD_ROLE;
import prompt.csv.utils.UnicodeBOMInputStream;
import prompt.csv.utils.context.DummyGlobalContext;
import prompt.csv.utils.context.DummyPluginContext;


public class CaseIdentifier {

	private GlobalContext context = new DummyGlobalContext();
	private PluginContext pluginContext = new DummyPluginContext(context, "");

	public HashMap<String, Double> InferCaseId(CSVLog csv, int timestampIndex, int activityIndex,
			int candidateCaseIdIndex) throws Exception {

		List<String> colsname = csv.getHeaders();

		String alg = "imfa";

		Map<TreeFitnessAbstract, Double> v = getValues(csv, timestampIndex, activityIndex, candidateCaseIdIndex, alg);

		HashMap<String, Double> metricsTEMP = new HashMap<String, Double>();

		for (TreeFitnessAbstract tfa : v.keySet()) {

			metricsTEMP.put(tfa.getInfo().toString(), v.get(tfa));
		}
	

		return metricsTEMP;

	}

	public int InferCaseIdFromLog(CSVLog csv, int timestampIndex, int activityIndex,
			ArrayList<Integer> candidateCaseIdIndexes,UIPluginContext context) throws Exception {

		HashMap<Integer, HashMap<String, Double>> data = new HashMap<Integer, HashMap<String, Double>>();

		double totalratios = 0;
       int counter = 0 ;
		for (int candidateCaseIdIndex : candidateCaseIdIndexes) {
			System.out.println(counter+" calculating quality metrics for candidate case: " + csv.getHeaders().get(candidateCaseIdIndex));
//			System.out.println("grouping ratio "+GroupingRate.GetGroupingScore(csv, candidateCaseIdIndex));
			context.getProgress().inc();

			HashMap<String, Double> entry = InferCaseId(csv, timestampIndex, activityIndex,
					candidateCaseIdIndex);
			entry.put("gpratio", GroupingRate.GetGroupingScore(csv, candidateCaseIdIndex).doubleValue());
			totalratios += entry.get("gpratio");
			data.put(candidateCaseIdIndex, entry);
			counter++;
		}
		
		/// add gpratio for event name and timestamp to total ratios
		totalratios += GroupingRate.GetGroupingScore(csv, timestampIndex).doubleValue();
		totalratios += GroupingRate.GetGroupingScore(csv, activityIndex).doubleValue();

		double averageRatio = totalratios / csv.getHeaders().size();
		//

		int c = -1;
		double best = 0;

		for (Integer condidateCaseIdIndex : data.keySet()) {
//		Iterator it = data.entrySet().iterator();
		
//		while (it.hasNext()) {
			
//			HashMap.Entry pair = (HashMap.Entry) it.next();

//			Integer condidateCaseIdIndex = (Integer) pair.getKey();
			/*Hash*/Map<String, Double> entry = /*(HashMap<String, Double>)*/ data.get(condidateCaseIdIndex); //pair.getValue();
			
			entry.put("gpratio", round(entry.get("gpratio"),4));
			double gr = round((1 - Math.abs(entry.get("gpratio") - averageRatio)),4);

			
			entry.put("Fr", round(entry.get("Fr"),4));
			entry.put("Pi", round(entry.get("Pi"),4));
			entry.put("Sm", round(entry.get("Sm"),4));
			entry.put("Gv", round(entry.get("Gv"),4));
			
			double qual = (entry.get("Fr") + entry.get("Pi") + entry.get("Sm") + entry.get("Gv")
			+ gr) / 5 ;  /// compute quality score
			
//			System.out.println("gr"+gr);
//			System.out.println("Fr"+entry.get("Fr"));
//			System.out.println("Pi"+entry.get("Pi"));
//			System.out.println("Sm"+entry.get("Sm"));
//			System.out.println("gv"+entry.get("Gv"));
//			System.out.println("Quality score"+qual);
			

			if (qual > best) {
				best = qual;
				c = condidateCaseIdIndex;
			}

		}

		
		return c;
	}

	public static double round(double value, int places) {
	    if (places < 0) throw new IllegalArgumentException();

	    long factor = (long) Math.pow(10, places);
	    value = value * factor;
	    long tmp = Math.round(value);
	    return (double) tmp / factor;
	}
	
	public void generateLogMetrics(String CSVFile, int timestampIndex, int activityIndex, int candidateCaseIdIndex,
			PrintStream out) throws Exception {

		CSVLog csv = read(CSVFile);
		ArrayList<String> colsname = getCandidateColumnsName(CSVFile);

		String[] algs = new String[] { "imfa"/* ,"ima","imf","im","imc" */ }; // ,"imc" "imfa","ima","imf","im"

		for (String alg : algs) {
			Map<TreeFitnessAbstract, Double> v = getValues(csv, timestampIndex, activityIndex, candidateCaseIdIndex,
					alg);

			HashMap<String, Double> metricsTEMP = new HashMap<String, Double>();

			for (TreeFitnessAbstract tfa : v.keySet()) {
				metricsTEMP.put(tfa.getInfo().toString(), v.get(tfa));
			}

			out.print("\"" + alg + "\"\t");
			out.print("\"" + CSVFile + "\"\t");
			out.print("\"" + colsname.get(candidateCaseIdIndex) + "\"\t");
			for (String m : new String[] { "Fr", "Pi", "Sm", "Gv" }) {
				out.print("\"" + m + "\"\t");
				out.print(metricsTEMP.get(m) + "\t");
			}
			out.println("");

		}
	}

	public ArrayList<String> getCandidateColumnsName(String filename) throws IOException {

		File file = new File(filename);
		FileReader fileReader = new FileReader(file);
		BufferedReader bufferedReader = new BufferedReader(fileReader);
		String line;
		line = bufferedReader.readLine().replace("\"", "");
		StringTokenizer st = new StringTokenizer(line, ",");
		ArrayList<String> colsname = new ArrayList<String>();

		while (st.hasMoreElements()) {
			colsname.add(st.nextToken());
		}

		fileReader.close();

		return colsname;

	}

	public static CSVLog read(String file) throws Exception {
		CSVLog log = new CSVLog(file);
		UnicodeBOMInputStream ubis = new UnicodeBOMInputStream(new FileInputStream(file));
		ubis.skipBOM();
		InputStreamReader isr = new InputStreamReader(ubis);
		BufferedReader reader = new BufferedReader(isr);
		String readLine;
		while ((readLine = reader.readLine()) != null) {
			log.addLine(readLine);
		}
		reader.close();
		isr.close();
		ubis.close();
		log.prepare();
		return log;
	}

	private Map<TreeFitnessAbstract, Double> getValues(CSVLog csv, int time, int activityName, int caseId, String alg) {

		XLog log = getXLog(csv, time, activityName, caseId);
		// {"im","ima","imf","imfa","imc"}
		ProcessTree pt = null;
		if (alg.equals("im"))
			pt = this.doInductiveMinerIM(log);
		else if (alg.equals("ima"))
			pt = this.doInductiveMinerIMa(log);
		else if (alg.equals("imf"))
			pt = this.doInductiveMinerIMf(log);
		else if (alg.equals("imfa"))
			pt = this.doInductiveMinerIMfa(log);
		else if (alg.equals("imc"))
			pt = this.doInductiveMinerIMc(log);

		CentralRegistry registry = new CentralRegistry(log, new XEventNameClassifier(), new Random(1));
		ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree tree = converter.convert(pt);

		List<TreeFitnessAbstract> evals = new ArrayList<TreeFitnessAbstract>();
		evals.add(new FitnessReplay(registry, new DummyCanceler())); // what is DummyCanceler ?

		//		evals.add(new PrecisionEscEdges(registry));
		evals.add(new PrecisionEscEdgesImproved(registry));
		//		evals.add(new PrecisionFlowerSize());
		//		evals.add(new PrecisionOperatorCosts());
		//		evals.add(new SimplicityDuplMissingAct(registry));
		evals.add(new SimplicityMixed());
		//		evals.add(new SimplicityOperatorAlternation());
		//		evals.add(new SimplicityTreeSize());
		//		evals.add(new SimplicityUselessNodes());
		evals.add(new Generalization(registry));
		//		evals.add(new GeneralizationByFitnessReplayDeviation(registry));

		Map<TreeFitnessAbstract, Double> values = new HashMap<TreeFitnessAbstract, Double>();

		for (TreeFitnessAbstract tfa : evals) {

			values.put(tfa, tfa.getFitness(tree, Arrays.asList(tree)));

			System.gc();
		}

		return values;
	}

	private ProcessTree doInductiveMinerIM(XLog log) {
		IM im = new IM();
		MiningParametersIM para = new MiningParametersIM();
		return im.mineProcessTree(pluginContext, log, para);
	}

	private ProcessTree doInductiveMinerIMf(XLog log) {
		IM im = new IM();
		MiningParametersIMf para = new MiningParametersIMf();

		return im.mineProcessTree(pluginContext, log, para);

	}

	private ProcessTree doInductiveMinerIMfa(XLog log) {
		IM im = new IM();
		MiningParametersIMfa para = new MiningParametersIMfa();

		return im.mineProcessTree(pluginContext, log, para);

	}

	private ProcessTree doInductiveMinerIMa(XLog log) {
		IM im = new IM();
		MiningParametersIMa para = new MiningParametersIMa();

		return im.mineProcessTree(pluginContext, log, para);

	}

	private ProcessTree doInductiveMinerIMc(XLog log) {
		IM im = new IM();
		MiningParametersIMc para = new MiningParametersIMc();

		return im.mineProcessTree(pluginContext, log, para);

	}

	private XLog getXLog(CSVLog csv, int time, int activityName, int caseId) {
		csv.setRole(time, FIELD_ROLE.DATE);
		csv.setRole(activityName, FIELD_ROLE.ACTIVITY_NAME);
		csv.setRole(caseId, FIELD_ROLE.CASE_ID);
		return csv.toXLog(null);
	}
}

class DummyCanceler implements Canceller {
	public boolean isCancelled() {
		return false;
	}
}
