package prompt.csv.plugin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.GlobalContext;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;

import nl.tue.astar.AStarThread.Canceller;
import prompt.csv.data.CSVLog;
import prompt.csv.data.FIELD_ROLE;
import prompt.csv.data.FIELD_TYPE;
import prompt.csv.data.ICIrequiredFields;
import prompt.csv.data.SortedEvent;
import prompt.csv.utils.context.DummyGlobalContext;
import prompt.csv.utils.context.DummyPluginContext;

public class CaseIdentifier {

	private GlobalContext context = new DummyGlobalContext();
	private PluginContext pluginContext = new DummyPluginContext(context, "");
	
	@Plugin(
		name = "Infer Case ID",
		parameterLabels = { "CSV file" },
		returnLabels = { "Labeled Event Log" },
		returnTypes = { CSVLog.class },
		userAccessible = true
	)
	@UITopiaVariant(
		affiliation = "Technical University of Denmark",
		author = " A. A. Andaloussi, A. Burattin, B. Weber",
		email = "amab@dtu.dk, andbur@dtu.dk, bweb@dtu.dk"
	)
	public CSVLog inferCaseId(UIPluginContext context, final CSVLog csv) throws Exception {
	
		
		
      
		
		
		context.getProgress().setIndeterminate(true);
		
		iciPlugin.CaseIdentifier ci = new iciPlugin.CaseIdentifier();
		
		
		ICIrequiredFields rfields = new ICIrequiredFields() ;
		
		PromptParameters  parameters;
		PromptParameters  parameters2;
		PromptParameters  parameters3;
		
		boolean cancel = false ;
		
	
		
		do {
	     parameters = new PromptParameters(rfields);
		InteractionResult result = context.showConfiguration("Select TimeStamp Attribute", parameters.visualize(context, csv,0));
		if (result.equals(InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
			cancel = true ;
		}
		

//		context.showWizard("test", true, true, parameters);

		if(result.equals(InteractionResult.CONTINUE) && !parameters.isSelection()) {
			 JOptionPane.showMessageDialog(new JFrame(), "No attribute is selected, please select the timestamp attribute.", "Error",
				        JOptionPane.ERROR_MESSAGE);
		}
		
		
		} while(!parameters.isSelection() && !cancel);
		
		 do {
				parameters2 = new PromptParameters(rfields);
				InteractionResult result2 = context.showConfiguration("Select Event Name Attribute", parameters2.visualize(context, csv,1));
				if (result2.equals(InteractionResult.CANCEL)) {
					context.getFutureResult(0).cancel(true);
					cancel = true ;
				}
				if(result2.equals(InteractionResult.CONTINUE) && !parameters2.isSelection()) {
					 JOptionPane.showMessageDialog(new JFrame(), "No attribute is selected, please select the event name attribute.", "Error",
						        JOptionPane.ERROR_MESSAGE);
				}
		        } while(!parameters2.isSelection() && !cancel);
		 
		 
		
				parameters3 = new PromptParameters(rfields);
				InteractionResult result3 = context.showConfiguration("Select Sampling Parameters", parameters3.visualize(context, csv,2));
				if (result3.equals(InteractionResult.CANCEL)) {
					context.getFutureResult(0).cancel(true);
					cancel = true ;
				}
		       
		 
		 
		 
		 
	
		 if(!cancel) {
			
			 
		final int timestampIndex = rfields.getTimestampIndex();
		csv.setType(timestampIndex, FIELD_TYPE.DATE);
        csv.setRole(timestampIndex, FIELD_ROLE.DATE);
		
       
		
		final int eventnameIndex = rfields.getEventnameIndex();
		csv.setType(eventnameIndex, FIELD_TYPE.STRING);
        csv.setRole(eventnameIndex, FIELD_ROLE.ACTIVITY_NAME);
	
        final int samplingSize = rfields.getSamplingSize();
        final int samplingMethod = rfields.getSamplingMethod();
		
		ArrayList<Integer> candidateCaseIdIndexes = new ArrayList<Integer>() {
			{
				for (String candidateCaseIdColumn : csv.getHeaders()) {
					int candidateCaseIdIndex = csv.getHeaders()
							.indexOf(candidateCaseIdColumn);
					if (candidateCaseIdIndex !=  timestampIndex
							&& candidateCaseIdIndex != eventnameIndex ) {
						add(candidateCaseIdIndex);
					}

				}
			}
		};
	
	context.getProgress().setIndeterminate(false);
	context.getProgress().setMinimum(0);
	context.getProgress().setMaximum(candidateCaseIdIndexes.size());
		
//	

//    System.out.println("params:"+ timestampIndex+", "+eventnameIndex+", "+samplingSize+", "+samplingMethod);
	
    int caseId = ci.InferCaseIdFromLog(compress(csv, samplingSize, samplingMethod, timestampIndex, rfields), timestampIndex, eventnameIndex,  candidateCaseIdIndexes,context);
    System.out.println("case id Index: "+caseId);

    csv.setType(caseId, FIELD_TYPE.STRING);
    csv.setRole(caseId, FIELD_ROLE.CASE_ID);
    
    context.getProgress().inc();
  
		 }
		
		 
		return  csv ;
	}
	

	public CSVLog sortCSV(CSVLog csv, int timeIndex) throws Exception {
		
		
		
		ArrayList<SortedEvent> eventsList = new ArrayList<SortedEvent>();
		
	for(int i=0 ; i<csv.getEntries(); i++) {
			
			List<String> event = csv.getLine(i);
			String timestamp = event.get(timeIndex);
			
			SortedEvent se = new SortedEvent(event,timestamp);
			eventsList.add(se);
	}
		
	   Collections.sort(eventsList);
	   
	   CSVLog sortedlog = new CSVLog("slog");
	   
	   List<String> headerL = csv.getHeaders();
		StringBuilder header = new StringBuilder();
		for(int ch = 0 ; ch<headerL.size(); ch++) {
			 header.append(headerL.get(ch));
			 if(ch<headerL.size()-1) {
				 header.append(",");
			 }
		}
	
	   sortedlog.addLine(header.toString());
	   
	   
	   for(SortedEvent ee : eventsList) {
		   List<String> entry  =ee.getEvent();
			StringBuilder event = new StringBuilder();
		   for(int ce = 0 ; ce<entry.size(); ce++) {
			   event.append(entry.get(ce));
			   if(ce<entry.size()-1) {
					 event.append(",");
				 }
			   
		   }
		   sortedlog.addLine(event.toString());
	   }
	   
	   sortedlog.prepare();

		return sortedlog ;
		
	}
	
	
	private CSVLog compress(CSVLog csv, int samplingSize, int samplingMethod, int timeIndex, ICIrequiredFields rfields) throws Exception {
		
		// check if full log
		if (rfields.isFullLog()) {
			System.out.println("Using full log");
			return csv;
		}
		
		 csv = sortCSV(csv,timeIndex);
		
		CSVLog selectedevents = new CSVLog("compressed log");
		
		
		int start = 0 ;
		int limit = 0  ;
		
		// check if sample size is lower than file size, otherwise the limit is the file size
		if(samplingSize<csv.getEntries()) {
			
			// check sampling method
			if(samplingMethod==0) {
				// top entries
				limit =  samplingSize ;
			}
			else if(samplingMethod==1) {
				// generate random start in the range  [0, filezie-samplingsize]
				Random rand = new Random();
				int min = 0 ;
				int max = csv.getEntries()-samplingSize ;
				start = rand.nextInt((max - min) + 1) + min;
				
				/// limit = start + sampling size
				limit = start + samplingSize;
				
			
			}
			
		}
		else {
			limit =  csv.getEntries() ;
		}
		
		System.out.println("sampling range["+start+","+limit+"]");
		
		// copy header
		StringBuilder header = new StringBuilder();
		for(String col : csv.getHeaders()) {
			header.append(col);
			header.append(',');
		}
		selectedevents.addLine(header.toString());
		
		// copy n entries
		for(int i=start; i<limit ; i++) {
			
			StringBuilder line = new StringBuilder();
			for(String col : csv.getLine(i)) {
				line.append(col);
				line.append(',');
			}
			
			selectedevents.addLine(line.toString());
			
		}
		
		selectedevents.prepare();
		return selectedevents ;
	}
	

class DummyCanceler implements Canceller {
	public boolean isCancelled() {
		return false;
	}
}
}
